local AddonName, fPB = ...
L = fPB.L


local	C_NamePlate_GetNamePlateForUnit, C_NamePlate_GetNamePlates, CreateFrame, CreateTexture, CreateFontString, UnitDebuff, UnitBuff, UnitIsUnit, UnitIsPlayer, UnitPlayerControlled, UnitIsEnemy, UnitIsFriend, GetSpellInfo, table_sort, strmatch, type, wipe, pairs, GetTime, math_floor =
		C_NamePlate.GetNamePlateForUnit, C_NamePlate.GetNamePlates, CreateFrame, CreateTexture, CreateFontString, UnitDebuff, UnitBuff, UnitIsUnit, UnitIsPlayer, UnitPlayerControlled, UnitIsEnemy, UnitIsFriend, GetSpellInfo, table.sort, strmatch, type, wipe, pairs, GetTime, math.floor

local defaultSpells1, defaultSpells2 = fPB.defaultSpells1, fPB.defaultSpells2

local LSM = LibStub("LibSharedMedia-3.0")
fPB.LSM = LSM

local config = LibStub("AceConfig-3.0")
local dialog = LibStub("AceConfigDialog-3.0")

fPB.db = {}
local db

local font = fPB.font

local fPBMainOptions
local fPBSpellsList
local fPBProfilesOptions


local PlatesBuffs = {}


local DefaultSettings = {
	profile = {
		showDebuffs = 2,		-- 1 = all, 2 = mine + spellList, 3 = only spellList, 4 = only mine
		showBuffs = 3,			-- 1 = all, 2 = mine + spellList, 3 = only spellList, 4 = only mine
		hidePermanent = true,

		showOnPlayers = true,
		showOnPets = true,
		showOnNPC = true,

		showOnEnemy = true,
		showOnFriend = true,
		showOnNeutral = true,

		showOnlyInCombat = false,
		showUnitInCombat = false,

		parentWorldFrame = false,
		
		baseWidth = 24,
		baseHeight = 24,
		myScale = 0.2,

		xInterval = 4,
		yInterval = 12,

		xOffset = 0,
		yOffset = 4,
		
		buffPerLine = 6,
		numLines = 3,

		showStdCooldown = true,
		showStdSwipe = false,
		
		showDuration = true,
		durationSize = 10,	
		blinkTimeleft = 0.2,
		
		stackSize = 10,
		
		font = "Friz Quadrata TT",

		borderStyle = 1,	-- 1 = \\texture\\border.tga, 2 = Blizzard, 3 = none
		colorizeBorder = true,
		colorTypes = {
			Magic 	= {0.20,0.60,1.00},
			Curse 	= {0.60,0.00,1.00},
			Disease = {0.60,0.40,0},
			Poison 	= {0.00,0.60,0},
			none 	= {0.80,0,   0},
			Buff 	= {0.00,1.00,0},
		},

		Spells = {},
		ignoredDefaultSpells = {},

		showSpellID = false,
	},
}


do --add default spells
for i=1, #defaultSpells1 do
	local spellID = defaultSpells1[i]
	local name = GetSpellInfo(spellID)
	if name then
		DefaultSettings.profile.Spells[name] = {
			spellID = spellID,
			scale = 2,
			durationSize = 18,
			show = 1,	-- 1 = always, 2 = mine, 3 = never
			stackSize = 18,
		}
	end
end

for i=1, #defaultSpells2 do
	local spellID = defaultSpells2[i]
	local name = GetSpellInfo(spellID)
	if name then
		DefaultSettings.profile.Spells[name] = {
			spellID = spellID,
			scale = 1.5,
			durationSize = 14,
			show = 1,	-- 1 = always, 2 = mine, 3 = never
			stackSize = 14,
		}
	end
end

end

local timeIntervals = {
	[1]	= {60 * 60 * 24 * 365, "y"},-- year
	[2]	= {60 * 60 * 24 * 30, "mo"},-- month
 	[3]	= {60 * 60 * 24 * 7, "w"},-- week
	[4]	= {60 * 60 * 24, "d"},-- day
	[5]	= {60 * 60, "h"},-- hour
	[6]	= {60, "m"},-- minute
}

local function FormatTime(seconds)
	local text = ""
	if seconds == 0 then
		text = "0s"
	else
		for i, data in ipairs(timeIntervals) do
			local num = 0		
			while seconds > data[1] do
				num	= num + 1;
				seconds	= seconds - data[1];
			end
			if num > 0 then
				text = num..data[2];
				return text
			end
		end
		return seconds
	end
end

local function GetColorByTime(current, max)
	if max == 0 then max = 1 end
	local percentage = (current/max)*100
	local red,green = 0,0
	if percentage >= 50 then
		--green to yellow
		green		= 1
		red			= ((100 - percentage) / 100) * 2
	else
		--yellow to red
		red	= 1
		green		= ((100 - (100 - percentage)) / 100) * 2
	end
	return red, green, 0
end

local total = 0
local function iconOnUpdate(self, elapsed)
	total = total + elapsed
	if total > 0 then
		total = 0
		if self.expiration and self.expiration > 0 then
			local timeLeft = self.expiration - GetTime()					
			if timeLeft < 10 then
				timeLeft = math_floor(timeLeft * 10 + 0.5)/10
			else
				timeLeft = math_floor(timeLeft)
			end
			if db.showDuration then
				self.cd:SetText(FormatTime(timeLeft))
				self.cd:SetTextColor(GetColorByTime(timeLeft,self.duration))
				self.cdbg:SetWidth(self.cd:GetStringWidth())				
				self.cdbg:SetHeight(self.cd:GetStringHeight())
			end
			if (timeLeft / (self.duration + 0.01) ) < db.blinkTimeleft and timeLeft < 60 then --buff only has 20% timeleft and is less then 60 seconds.
				local f = GetTime() % 1
				if f > 0.5 then 
					f = 1 - f
				end
				self:SetAlpha(f * 3)
			end
		end
	end
end


local function DrawOnPlate(frame)
	
	local pBuffs = PlatesBuffs[frame]

	--[[if parent == frame then it will change scale and alpha with nameplates
	otherwise use UIParent, but this causes mess of icon/border textures]]
	if db.parentWorldFrame then
		frame.fPBiconsFrame = CreateFrame("Frame", nil, WorldFrame)
	else
		frame.fPBiconsFrame = CreateFrame("Frame", nil, frame)
	end
	
	local sumWidth = {}
	local maxHeight = {}
	local sumHeight = 0
	
	frame.fPBiconsFrame.linesFrame = {}
	local buffLine = frame.fPBiconsFrame.linesFrame
	frame.fPBiconsFrame.iconsFrame = {}
	local buffIcon = frame.fPBiconsFrame.iconsFrame
	
	local breaked = false
	for l = 1, db.numLines do
		if breaked then break end

		buffLine[l] = CreateFrame("Frame", nil, frame.fPBiconsFrame)
		if l == 1 then
			buffLine[l]:SetPoint("BOTTOMLEFT", frame.fPBiconsFrame, "BOTTOMLEFT", 0, 0 )
		else
			buffLine[l]:SetPoint("BOTTOMLEFT", buffLine[l-1], "TOPLEFT", 0, db.yInterval )
		end

		sumWidth[l] = 0
		maxHeight[l] = 0

		for k = 1, db.buffPerLine do
			local i = db.buffPerLine*(l-1)+k
			
			buffIcon[i] = CreateFrame("Frame", nil, buffLine[l])

			local scale = (pBuffs[i].my and db.myScale + 1 or 1) * (pBuffs[i].scale or 1)
			local width = db.baseWidth * scale
			sumWidth[l] = sumWidth[l] + width + db.xInterval

			local height = db.baseHeight * scale
			maxHeight[l] = (height > maxHeight[l]) and height or maxHeight[l]

			buffIcon[i]:SetWidth(width)
			buffIcon[i]:SetHeight(height)

			if k == 1 then
				buffIcon[i]:SetPoint("BOTTOMLEFT", buffLine[l], "BOTTOMLEFT", 0, 0)
			else
				buffIcon[i]:SetPoint("BOTTOMLEFT", buffIcon[i-1], "BOTTOMRIGHT", db.xInterval, 0)
			end

			local texture = buffIcon[i]:CreateTexture(nil, "BACKGROUND")
			texture:SetTexture(pBuffs[i].icon)
			texture:SetAllPoints(buffIcon[i])

			if pBuffs[i].stack > 0 then
				texture.stack = buffIcon[i]:CreateFontString(nil, "ARTWORK", "ChatFontNormal")
				texture.stack:SetText(tostring(pBuffs[i].stack))
				texture.stack:SetFont(fPB.font, (pBuffs[i].stackSize or db.stackSize), "OUTLINE")
				texture.stack:SetPoint("BOTTOMRIGHT", texture, "BOTTOMRIGHT", -1, 3)
			end
			
			if db.borderStyle ~= 3 then
				texture.border = buffIcon[i]:CreateTexture(nil,"BORDER")

				if db.borderStyle == 1 then
					texture.border:SetTexture("Interface\\Addons\\flyPlateBuffs\\texture\\border.tga")
					texture.border:SetTexCoord(0.08,0.08, 0.08,0.92, 0.92,0.08, 0.92,0.92)		--хз почему отображает не на всю иконку по дефолту, цифры подбором
				elseif db.borderStyle == 2 then
					texture.border:SetTexture("Interface\\Buttons\\UI-Debuff-Overlays")
					texture.border:SetTexCoord(0.296875,0.5703125,0,0.515625)		-- if "Interface\\Buttons\\UI-Debuff-Overlays"
				end
				local color
				if pBuffs[i].type == "buff" then
					color = db.colorTypes.Buff
				else
					if db.colorizeBorder then
						color = db.colorTypes[pBuffs[i].debufftype] or db.colorTypes.none
					else
						color = db.colorTypes.none
					end														
				end
				texture.border:SetVertexColor(color[1], color[2], color[3])
				texture.border:SetAllPoints(texture)
			end

			if (db.showStdCooldown or db.showStdSwipe) and pBuffs[i].expiration > 0 then
				texture.cooldown = CreateFrame("Cooldown", nil, buffIcon[i], "CooldownFrameTemplate")
				if not db.showStdSwipe then
					texture.cooldown:SetDrawSwipe(false)
				else
					texture.cooldown:SetSwipeColor(0, 0, 0, 0.6)
				end
				if not db.showStdCooldown then
					texture.cooldown:SetHideCountdownNumbers(true)
				end
				texture.cooldown:SetReverse(true)
				texture.cooldown:SetDrawEdge(false)
				texture.cooldown:SetCooldown(pBuffs[i].expiration - pBuffs[i].duration, pBuffs[i].duration)
			end

			if db.showDuration and pBuffs[i].expiration > 0 then
				buffIcon[i].cd = buffIcon[i]:CreateFontString(nil, "ARTWORK", "ChatFontNormal")
				buffIcon[i].cd:SetPoint("TOP", buffIcon[i], "BOTTOM", 0, -1)
				buffIcon[i].cd:SetFont(fPB.font, (pBuffs[i].durationSize or db.durationSize), "NORMAL")
				
				buffIcon[i].cdbg = buffIcon[i]:CreateTexture(nil,"BACKGROUND")
				buffIcon[i].cdbg:SetColorTexture(0,0,0,.75)
				buffIcon[i].cdbg:SetPoint("CENTER", buffIcon[i].cd)
			end	
			
			if (db.showDuration or db.blinkTimeleft > 0) and pBuffs[i].expiration > 0 then
				buffIcon[i].expiration = pBuffs[i].expiration
				buffIcon[i].duration = pBuffs[i].duration
				buffIcon[i]:SetScript("OnUpdate",iconOnUpdate)
			end
			
			if i == #pBuffs then breaked = true; break end -- if no more buffs
		end
		sumHeight = sumHeight + maxHeight[l] + db.yInterval
		buffLine[l]:SetWidth(sumWidth[l]-db.xInterval)
		buffLine[l]:SetHeight(maxHeight[l])
	end
	table_sort(sumWidth) -- max width
	frame.fPBiconsFrame:SetWidth(sumWidth[#sumWidth]-db.xInterval)
	frame.fPBiconsFrame:SetHeight(sumHeight - db.yInterval)
	frame.fPBiconsFrame:SetPoint("BOTTOM",frame,"TOP",db.xOffset,db.yOffset)
	-- local testTexture = frame.fPBiconsFrame:CreateTexture(nil,"BACKGROUND")
	-- testTexture:SetColorTexture(0.5, 0.5, 0.5)
	-- testTexture:SetAllPoints(frame.fPBiconsFrame)

	frame.fPBiconsFrame:Show()

end

local function FilterBuffs(frame, type, name, _, icon, stack, debufftype, duration, expiration, caster, _, _, spellID)
	local Aura = {
		type = type,
		name = name,
		icon = icon,
		stack = stack,
		debufftype = debufftype,
		duration = duration,
		expiration = expiration,
		spellID = spellID,
		my = caster == "player",
	}
	local pBuffs = PlatesBuffs[frame]
	local listedSpell
	if (db.Spells[name] and not db.ignoredDefaultSpells[name]) and not (db.Spells[name].checkID and db.Spells[name].spellID ~= spellID)  then	
		listedSpell = db.Spells[name]
		Aura.scale = listedSpell.scale
		Aura.durationSize = listedSpell.durationSize
		Aura.stackSize = listedSpell.listedSpell
	end
	if db.hidePermanent and Aura.duration == 0 and not (listedSpell and (listedSpell.show == 1 or listedSpell.show == 2)) then
		return
	end

	-- showDebuffs  1 = all, 2 = mine + spellList, 3 = only spellList, 4 = only mine
	-- listedSpell.show  -- 1 = always, 2 = mine, 3 = never
	if not (listedSpell and listedSpell.show == 3) then
		if type == "debuff" then
			if db.showDebuffs == 1 then
				pBuffs[#pBuffs + 1] = Aura
				return
			elseif (db.showDebuffs == 2 or db.showDebuffs == 4) and Aura.my then
				pBuffs[#pBuffs + 1] = Aura
				return
			elseif listedSpell and (db.showDebuffs == 2 or db.showDebuffs == 3) then
				if listedSpell.show == 1 then
					pBuffs[#pBuffs + 1] = Aura
					return
				elseif listedSpell.show == 2 and Aura.my then
					pBuffs[#pBuffs + 1] = Aura
					return
				else
					return
				end
			else
				return
			end
		elseif type == "buff" then
			if db.showBuffs == 1 then
				pBuffs[#pBuffs + 1] = Aura
				return
			elseif (db.showBuffs == 2 or db.showBuffs == 4) and Aura.my then
				pBuffs[#pBuffs + 1] = Aura
				return
			elseif listedSpell and (db.showBuffs == 2 or db.showBuffs == 3) then
				if listedSpell.show == 1 then
					pBuffs[#pBuffs + 1] = Aura
					return
				elseif listedSpell.show == 2 and Aura.my then
					pBuffs[#pBuffs + 1] = Aura
					return
				else
					return
				end
			else
				return
			end
		end
	end
end


local function ScanUnitBuffs(nameplateID, frame)
	
	if not PlatesBuffs[frame] then
		PlatesBuffs[frame] = {}
	end
	
	for i = 1, 40 do
		local name, _, icon, stack, debufftype, duration, expiration, caster, _, _, spellID = UnitDebuff(nameplateID,i)
		if name then
			FilterBuffs(frame, "debuff", name, _, icon, stack, debufftype, duration, expiration, caster, _, _, spellID)
		end
		local name, _, icon, stack, debufftype, duration, expiration, caster, _, _, spellID = UnitBuff(nameplateID,i)
		if name then
			FilterBuffs(frame, "buff", name, _, icon, stack, debufftype, duration, expiration, caster, _, _, spellID)
		end
	end
	
	if #PlatesBuffs[frame] > 0 then
		table_sort(PlatesBuffs[frame],function(a,b)
			if (a and b) then
				if a.type ~= b.type then -- debuffs before buffs
					return a.type > b.type
				elseif a.expiration == 0 then -- without duration last
					return false
				elseif b.expiration == 0 then
					return true
				-- elseif (a.scale or 1) ~= (b.scale or 1) then -- priority
					-- return (a.scale or 1) > (b.scale or 1)
				elseif a.my ~= b.my then -- self first
					return (a.my and 1 or 0) > (b.my and 1 or 0)
				elseif a.expiration == b.expiration then
					return a.name < b.name
				else
					return a.expiration < b.expiration
				end
			end
		end)

		DrawOnPlate(frame)
	end
end


local function ClearUnitBuffs(frame)
	if frame.fPBiconsFrame then
		frame.fPBiconsFrame:Hide()
		frame.fPBiconsFrame = nil
	end
	if PlatesBuffs[frame] then				
		wipe(PlatesBuffs[frame])
	end
end

local function UpdateUnitAuras(...)
	local nameplateID = ...
	local frame = C_NamePlate_GetNamePlateForUnit(nameplateID)
	ClearUnitBuffs(frame)
	
	if db.showOnlyInCombat and not UnitAffectingCombat("player") then return end -- InCombatLockdown()
	if db.showUnitInCombat and not UnitAffectingCombat(nameplateID) then return end
	
	-- filter units
	if UnitIsUnit(nameplateID,"player") then return end
	if UnitIsPlayer(nameplateID) and not db.showOnPlayers then return end
	if UnitPlayerControlled(nameplateID) and not UnitIsPlayer(nameplateID) and not db.showOnPets then return end
	if not UnitPlayerControlled(nameplateID) and not UnitIsPlayer(nameplateID) and not db.showOnNPC then return end
	if UnitIsEnemy(nameplateID,"player") and not db.showOnEnemy then return end
	if UnitIsFriend(nameplateID,"player") and not db.showOnFriend then return end
	if not UnitIsFriend(nameplateID,"player") and not UnitIsEnemy(nameplateID,"player") and not db.showOnNeutral then return end
	
	ScanUnitBuffs(nameplateID,frame)
	
end

function fPB.UpdateAllNameplates()
	for i, p in ipairs(C_NamePlate_GetNamePlates()) do
		UpdateUnitAuras(p.namePlateUnitToken)
	end
end
local UpdateAllNameplates = fPB.UpdateAllNameplates

local function Nameplate_Added(...)
	local nameplateID = ...
	local frame = C_NamePlate_GetNamePlateForUnit(nameplateID)
	
	frame.UnitFrame.BuffFrame:Hide()	--Hide terrible standart debuff frame
	
	if db.showOnlyInCombat and not InCombatLockdown() then return end
	
	UpdateUnitAuras(nameplateID)
end
local function Nameplate_Removed(...)
	local nameplateID = ...
	local frame = C_NamePlate_GetNamePlateForUnit(nameplateID)
	
	ClearUnitBuffs(frame)
	if PlatesBuffs[frame] then
		PlatesBuffs[frame] = nil
	end
end

function fPB.AddNewSpell(spellName, spellID)
	if db.ignoredDefaultSpells[spellName] then
		db.ignoredDefaultSpells[spellName] = nil
	end
	db.Spells[spellName] = {
		show = 1,
		spellID = spellID,
		scale = 1,
		stackSize = db.stackSize,
		durationSize = db.durationSize,
	}
	UpdateAllNameplates()
	fPB.BuildSpellList()
end
function fPB.RemoveSpell(spellName)
	if DefaultSettings.profile.Spells[spellName] then
		db.ignoredDefaultSpells[spellName] = true
	end
	db.Spells[spellName] = nil
	UpdateAllNameplates()
	fPB.BuildSpellList()
end

function fPB.OnProfileChanged()
	db = fPB.db.profile
	fPB.OptionsOnEnable()
	UpdateAllNameplates()
	fPB.BuildSpellList()
end
local function Initialize()
	fPB.db = LibStub("AceDB-3.0"):New("flyPlateBuffsDB", DefaultSettings, true)
	fPB.db.RegisterCallback(fPB, "OnProfileChanged", "OnProfileChanged")
	fPB.db.RegisterCallback(fPB, "OnProfileCopied", "OnProfileChanged")
	fPB.db.RegisterCallback(fPB, "OnProfileReset", "OnProfileChanged")
	
	db = fPB.db.profile
	
	fPB.font = fPB.LSM:Fetch("font", db.font)
	
	config:RegisterOptionsTable(AddonName, fPB.MainOptionTable)
	fPBMainOptions = dialog:AddToBlizOptions(AddonName, AddonName)
	
	config:RegisterOptionsTable(AddonName.." Spells", fPB.SpellsTable)
	fPBSpellsList = dialog:AddToBlizOptions(AddonName.." Spells", L["Specific spells"], AddonName)
	
	config:RegisterOptionsTable(AddonName.." Profiles", LibStub("AceDBOptions-3.0"):GetOptionsTable(fPB.db))
	fPBProfilesOptions = dialog:AddToBlizOptions(AddonName.." Profiles", L["Profiles"], AddonName)
	
	SLASH_FLYPLATEBUFFS1, SLASH_FLYPLATEBUFFS2 = "/fpb", "/pb"
	function SlashCmdList.FLYPLATEBUFFS(msg, editBox)
		InterfaceOptionsFrame_OpenToCategory(fPBMainOptions)
		InterfaceOptionsFrame_OpenToCategory(fPBSpellsList)
		InterfaceOptionsFrame_OpenToCategory(fPBMainOptions)
	end
end

function fPB.RegisterCombat()
	fPB.Events:RegisterEvent("PLAYER_REGEN_DISABLED")
	fPB.Events:RegisterEvent("PLAYER_REGEN_ENABLED")
end
function fPB.UnregisterCombat()
	fPB.Events:UnregisterEvent("PLAYER_REGEN_DISABLED")
	fPB.Events:UnregisterEvent("PLAYER_REGEN_ENABLED")
end

fPB.Events = CreateFrame("Frame")
fPB.Events:RegisterEvent("ADDON_LOADED")
fPB.Events:RegisterEvent("PLAYER_ENTERING_WORLD")

fPB.Events:SetScript("OnEvent", function(self, event, ...)
	if event == "ADDON_LOADED" and (...) == AddonName then
		Initialize()
	elseif event == "PLAYER_ENTERING_WORLD" then		
		fPB.OptionsOnEnable()
		fPB.FixBlizzard()
		if db.showSpellID then fPB.ShowSpellID() end
		
		fPB.Events:RegisterEvent("NAME_PLATE_UNIT_ADDED")
		fPB.Events:RegisterEvent("NAME_PLATE_UNIT_REMOVED")

		if db.showOnlyInCombat then
			fPB.RegisterCombat()
		else
			fPB.Events:RegisterEvent("UNIT_AURA")
		end
	elseif event == "PLAYER_REGEN_DISABLED" then
		fPB.Events:RegisterEvent("UNIT_AURA")
		UpdateAllNameplates()
	elseif event == "PLAYER_REGEN_ENABLED" then
		fPB.Events:UnregisterEvent("UNIT_AURA")
		UpdateAllNameplates()
	elseif event == "NAME_PLATE_UNIT_ADDED" then
		Nameplate_Added(...)
	elseif event == "NAME_PLATE_UNIT_REMOVED" then
		Nameplate_Removed(...)
	elseif event == "UNIT_AURA" then
		if strmatch((...),"nameplate%d+") then
			UpdateUnitAuras(...)
		end
	end
end)
